/* simple program for BrickOS (version 0.9) that sends an IR message from RCX to WowWee robots */

/* PC -> RCX protocol (LNP integrity messages):
 *
 *  all messages are 3-byte
 *
 *  immediate action:
 *    0x7D b1 b2      (where b1 is LSB of hexcode, and b2 is MSB of hexcode)
 *
 *  program a sequence:
 *    0x7E b1 b2      start of message, b1 - sequence bank number 0-15, b2 unused
 *    0x7D b1 b2      hexcode for action (as above) - the time when the message is sent is recorded for replay
 *    0x7A b1 b2      overwrite the time with a specified value (b1 is LSB), value in milliseconds (7A should follow after 7D)
 *    0x7C b1 b2      end of message, replay immediatelly, b1 - seq. bank number, b2 unused, or:
 *    0x7B b1 b2      end of message, store it for replay later, b1,b2 - unused
 *                          PRGM button starts sequence in current bank (even multiple times)
 *                          current bank for PRGM button is set to the bank just programmed
 *    0x79 b1 b2      select current bank (has effect only on PRGM button), b1 - seq. bank number, b2 unused
 *
 *    0x7D b1 b2      operate motor(s). b2 specifies the motors to work on (sum of the motor codes + 0x80)
 *                    and the direction the motors should take (motors not specified are not affected).
 *                           0x01 - motor A
 *                           0x02 - motor B
 *                           0x04 - motor C
 *                           0x00 - off
 *                           0x08 - forward
 *                           0x10 - reverse
 *                           0x18 - brake (stop and hold - eats power, should be followed by off)
 *                       b1 specifies the new motor speed (0x00 - 0xFF)
 */


#include <conio.h>
#include <unistd.h>
#include <dmotor.h>
#include <dbutton.h>
#include <lnp/lnp.h>
#include <lnp/lnp-logical.h>

#include <sys/h8.h>
#include <sys/time.h>

#define NUM_SEQ_BANKS 16
#define SEQUENCE_BANK_SIZE 100

inline void wait_unit()
{
  unsigned char b, c;

  for (b = 0; b < 255; b++)
    for (c = 0; c < 1; c++);
  for (b = 0; b < 25; b++)
    for (c = 0; c < 1; c++);
}

void sapien_send(unsigned int code)
{
  unsigned int x = (code < 0x100)?128:2048;

  asm("ORC #128, ccr");  /* clear interrupts */

  /* clear TE and RE bits */
  S_CR = S_CR & 0xCF;

  /* set direction on the P50 (TxD) to out */
  PORT5_DDR = 5;

  /* header */

  PORT5 = 0;
  /* wait 8/1200 sec */
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  wait_unit();
  PORT5 = 1;

  while (x)
  {
    if (code & x) /* wait 4/1200 sec */
    {
      wait_unit();
      wait_unit();
      wait_unit();
      wait_unit();
    }
    else /* wait 1/1200 sec */
      wait_unit();

    PORT5 = 0;
    /* wait 1/1200 sec */
    wait_unit();
    PORT5 = 1;

    /* move to next bit of the hexcode */
    x >>= 1;
  }

  /* set again TE and RE bits */
  S_CR = S_CR | 0x30;

  asm("ANDC #127, ccr"); /* set interrupts */
}

static unsigned char message;  /* message arrived (1 = simple, 2 = sequence, send now, 3 = sequence, on button) */
static unsigned int code;      /* code of the simple message */
static int si[NUM_SEQ_BANKS];             /* length of sequence message */
static unsigned int seq[NUM_SEQ_BANKS][SEQUENCE_BANK_SIZE + 1];    /* sequence message codes */
static time_t timing[NUM_SEQ_BANKS][SEQUENCE_BANK_SIZE + 1];       /* sequence message timing */
static int current_bank;      /* which sequence is replayed by PRGM button */
static int replay_bank;             /* currently replayed bank */
static int store_bank;              /* currently programmed bank */


void lnp_integrity_message(const unsigned char *buf, unsigned char ln)
{
  char ctrl;

  if (ln > 2)
  {
    ctrl = (char)buf[0];
    if (ctrl == 0x7E)  /* 7E: sequence start */
    {
      store_bank = (int)buf[1];
      if (store_bank > NUM_SEQ_BANKS) store_bank = 0;
      si[store_bank] = 0;
      timing[store_bank][0] = get_system_up_time();
      message = 0xFF;
    }
    if (ctrl == 0x7D)  /* 7D: sequence data */
    {
      if (message == 0xFF)
      {
        seq[store_bank][si[store_bank]++] = (unsigned int)buf[1] + ((unsigned int)buf[2] << 8);
        if (si[store_bank] == SEQUENCE_BANK_SIZE) si[store_bank] = 0;
        timing[store_bank][si[store_bank]] = get_system_up_time();
      }
      else  /* immediate message (was not preceded by sequence start) */
      {
        code = (unsigned int)buf[1] + ((unsigned int)buf[2] << 8);
        message = 1;
      }
    }
    if (ctrl == 0x7C)  /* 7C: sequence end - send immediatelly */
    {
      replay_bank = (int)buf[1];
      message = 2;
    }
    else if (ctrl == 0x7B) /* 7B: sequence end - do not send, wait for PRGM button */
    {
      current_bank = store_bank;
      message = 3;
    }
    else if (ctrl == 0x7A) /* 7A: overwrite time */
    {
      if (si[store_bank] > 0) timing[store_bank][si[store_bank]] = timing[store_bank][si[store_bank] - 1] + (int)buf[1] + ((int)buf[2] << 8);
    }
    else if (ctrl == 0x79) /* 79: select current bank */
      current_bank = (int)buf[1];
  }
}

void motor_command(unsigned int code)
{
  unsigned char op = (unsigned char)(code >> 8);
  unsigned char speed = (unsigned char)(code & 0xFF);
  MotorDirection mdir = (MotorDirection)((op >> 3) & 7);
  
  if (op & 0x01)
  {
    motor_a_speed(speed);
    motor_a_dir(mdir);
  }
  if (op & 0x02)
  {
    motor_b_speed(speed);
    motor_b_dir(mdir);
  }
  if (op & 0x04)
  {
    motor_c_speed(speed);
    motor_c_dir(mdir);
  }
}

int main(int argc, char **argv)
{
  time_t t0;
  unsigned int i, x;

  message = 0;
  current_bank = 0;
  for (i = 0; i < NUM_SEQ_BANKS; i++) si[i] = 0;
  lnp_integrity_handler = &lnp_integrity_message;
  lnp_logical_range(1);

  //systime_set_timeslice(255);
  cputs("ready");

  while (RELEASED(dbutton(),BUTTON_VIEW))
  {
    if (PRESSED(dbutton(), BUTTON_PROGRAM))
    {
      msleep(200);
      while (PRESSED(dbutton(), BUTTON_PROGRAM));
      msleep(50);
      replay_bank = current_bank;
      message = 2;
    }
    if (message == 1)
    {
      message = 0;
      if (code >> 15) motor_command(code);
      else sapien_send(code);
      //cputw(code);
    }
    else if (message == 2)
    {
      //cputs("work ");
      message = 0;
      t0 = get_system_up_time() - timing[replay_bank][0];
      i = 1;
      while (i <= si[replay_bank])
      {
	while (get_system_up_time() < timing[replay_bank][i] + t0);
	x = seq[replay_bank][i - 1];
	if (x >> 15) motor_command(x);
	else sapien_send(x);
	//cputw(i);
	i++;
      }
      //cputs("sent ");
    }
    else if (message == 3)
    {
      //cputs("progm");
      message = 0;
    }
    else if (message == 0xFF)
    {
      //cputw(si[store_bank]);
    }
  }

  cputs("done");
  sleep(1);
  cls();
  lnp_integrity_handler = LNP_DUMMY_INTEGRITY;

  return 0;
}

